/**
 * @file main.cpp
 * @brief 
 */

#include <iostream>
#include <fstream>
#include <stdexcept>
#include "cpxmacro.h"
#include "knapsack.h"

using namespace std;

const double EPS = 1e-5;

// error status and messagge buffer
int status;
char errmsg[BUF_SIZE];

// MKPData

class MKP
{
public:
	MKP() : m(0), n(0) {}
	int m;
	int n;
	std::vector<double> capacities;
	std::vector<double> profits;
	std::vector< std::vector<double> > weights;
	void read(const char* filename)
	{
		std::ifstream in(filename);
		// read size
		in >> m;
		in >> n;
		std::cout << "m = " << m << std::endl;
		std::cout << "n = " << n << std::endl;
		// read capacities
		for (int i = 0; i < m; i++)
		{
			double cap;
			in >> cap;
			capacities.push_back(cap);
		}
		// read profits
		for (int i = 0; i < n; i++)
		{
			double p;
			in >> p;
			profits.push_back(p);
		}
		// read weights
		weights.resize(m);
		for (int i = 0; i < m; i++)
		{
			weights[i].reserve(n);
			for (int j = 0; j < n; j++)
			{
				double w;
				in >> w;
				weights[i].push_back(w);
			}
		}
		in.close();
	}
};

// cut callback data
class CBData
{
public:
	CBData(MKP* _mkp, KPSolver* _solver) : mkp(_mkp), solver(_solver) {}
	MKP* mkp;
	KPSolver* solver;
};

int myCutCallback(CEnv env, void* cbdata, int wherefrom, void* cbhandle, int* useraction_p)
{
	// check if solution is fractional
	*useraction_p = CPX_CALLBACK_DEFAULT; ///tra le altre possibilita' CPX_CALLBACK_FAIL (per dirgli di finire perche' abbiamo avuto problemi)

	/// cplex chiama la callback anche se la soluzione e' intera (per permettere di implementare generazione di vincoli, come ad es. TSP)
	/// verifichiamo quindi che la soluzione sia effettivamente frazionaria (ce ne accorgeremmo, visto come separiamo, ma cosi' risparmiamo tempo)
	int numInt = 0;
	CHECKED_CPX_CALL( CPXgetcallbacknodeinfo, env, cbdata, wherefrom, 0, CPX_CALLBACK_INFO_NODE_NIINF, &numInt );
				/// 0 e' il nodo corrente
				/// NIIF: number of integer infeasibility
	if (numInt == 0) return 0;
	
	// get fractional solution
	CBData* mydata = static_cast<CBData*>(cbhandle);
	int n = mydata->mkp->n;
	std::vector<double> x(n);
	CHECKED_CPX_CALL( CPXgetcallbacknodex, env, cbdata, wherefrom, &x[0], 0, n - 1 );
	
	// separate
	int m = mydata->mkp->m;
	std::vector<double> obj(n);
	for (int j = 0; j < n; j++) obj[j] = 1.0 - x[j];
				/// la funzione obiettivo e' comune a tutti i problemi di separazione, qualsiasi sia il vincolo sul quale cerchero' la cover violata
	std::vector<double> y(n);
	double objval;
	for (int i = 0; i < m; i++)
	{
		mydata->solver->exec(mydata->mkp->weights[i], obj, mydata->mkp->capacities[i] + 1.0, y, objval);
				/// 1.0 e' l'epsilon per il >= W+epsilon: 1.0 perche' assumiamo coefficienti tutti interi.
		if (objval < (1.0 - EPS))
		{
			// we have a violated cover
			std::vector<int> idx;
			std::vector<double> coef;
			double rhs = -1.0;
			for(int j = 0; j < n; j++)
			{
				if (y[j] > 0.5)
				{
					idx.push_back(j);
					coef.push_back(1.0);
					rhs += 1.0;
				}
			}
			CHECKED_CPX_CALL( CPXcutcallbackadd, env, cbdata, wherefrom, idx.size(), rhs, 'L', &idx[0], &coef[0], 0 );
			std::cout << "Added 1 cover" << std::endl;
		}
	}
	*useraction_p = CPX_CALLBACK_SET;
				/// ho fatto qualcosa! altrimenti il taglio viene ignorato...
	return 0;
}

int main (int argc, char const *argv[])
{
	try
	{
		if (argc < 2) throw std::runtime_error("usage: ./main filename.dat [no_cover]");
		bool separateCoverCuts = (argc > 2);
		// init
		DECL_ENV( env );
		DECL_ENV( env2 );
		DECL_PROB( env, lp );
		MKP data;
		data.read(argv[1]);
		KPSolver kp(env2);
		
		// setup LP
		// add vars
		std::vector<char> xtype(data.n, 'B');
		CHECKED_CPX_CALL( CPXnewcols, env, lp, data.n, &(data.profits[0]), 0, 0, &xtype[0], 0 );
		CPXchgobjsen(env, lp, CPX_MAX);
		// add constraints
		int matbeg = 0;
		char sense = 'L';
		std::vector<int> idx(data.n);
		for (int i = 0; i < data.n; i++) idx[i] = i;
		for (int i = 0; i < data.m; i++)
		{
			CHECKED_CPX_CALL( CPXaddrows, env, lp, 0, 1, data.n, &(data.capacities[i]), &sense, &matbeg, &idx[0], &(data.weights[i][0]), 0, 0 );
		}
		CHECKED_CPX_CALL( CPXwriteprob, env, lp, "mkp.lp", 0 );
		
		// set appropriate parameters
		CHECKED_CPX_CALL( CPXsetintparam, env, CPX_PARAM_CUTPASS, -1 );
		      //upper limit on the number of cutting plane passes CPLEX performs when solving the root node of a MIP model
		CHECKED_CPX_CALL( CPXsetintparam, env, CPX_PARAM_HEURFREQ, -1 );
		      //how often (noe interval) to apply the periodic heuristic
		CHECKED_CPX_CALL( CPXsetintparam, env, CPX_PARAM_PREIND, CPX_OFF );
		      //apply presolver to simplify and reduce problems before the problem is solved
		CHECKED_CPX_CALL( CPXsetintparam, env, CPX_PARAM_MIPCBREDLP, CPX_OFF );
		      //access node information: original model (off) or reduced, presolved model (on, default)
		CHECKED_CPX_CALL( CPXsetintparam, env, CPX_PARAM_MIPDISPLAY, 4);
		CHECKED_CPX_CALL( CPXsetintparam, env, CPX_PARAM_SCRIND, CPX_ON);
		CHECKED_CPX_CALL( CPXsetintparam, env, CPX_PARAM_MIPINTERVAL, 1);
		      //log mip process info
		
		// set cut callback
		if (separateCoverCuts)
		{
			CBData mydata(&data, &kp);
			CHECKED_CPX_CALL( CPXsetcutcallbackfunc, env, &myCutCallback, (void*)&mydata );
		}
		
		// solve
		CHECKED_CPX_CALL( CPXmipopt, env, lp );
		double objval;
		CHECKED_CPX_CALL( CPXgetobjval, env, lp, &objval );
		std::cout << "Objval: " << objval << std::endl;
		std::cout << "Nodes: " << CPXgetnodecnt(env, lp) << std::endl;
		
		// free
		CPXfreeprob(env, &lp);
		CPXcloseCPLEX(&env);
		CPXcloseCPLEX(&env2);
		
	}
	catch(std::exception& e)
	{
		std::cout << ">>>EXCEPTION: " << e.what() << std::endl;
	}
	return 0;
}

